@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2002  The PARI group.
+Copyright (C) 2000-2013  The PARI group.
 
 This file is part of the GP2C package.
 
@@ -10,7 +10,7 @@ ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
 with the package; see the file 'COPYING'. If not, write to the Free Software
-Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.*/
 
 #include "config.h"
 #include <stdlib.h>
@@ -19,6 +19,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 #include <ctype.h>
 #include "header.h"
 int indent=0;
+extern int opthashline;
 extern int indentlevel;
 
 void genindent(FILE *fout)
@@ -28,10 +29,23 @@ void genindent(FILE *fout)
     fputc(' ',fout);
 }
 
+void genindentline(FILE *fout, char *s)
+{
+  genindent(fout);
+  fputs(s,fout);
+}
+
+void genlineno(FILE *fout, int n)
+{
+  if (opthashline)
+    fprintf(fout,"#line %d \"%s\"\n",tree[n].lineno,nameparse);
+}
+
 void genindentseq(FILE *fout, int n)
 {
-  if (n==-1 || (tree[n].m&(1<<Msemicomma)))
+  if (n==-1 || (tree[n].m&(1<<Msemicolon)))
     return;
+  genlineno(fout,n);
   genindent(fout);
 }
 
@@ -42,11 +56,12 @@ void gencomment(FILE *fout, int n, int flag)
   if (tree[n].comment<0)
     return;
   c=com+tree[n].comment;
+  tree[n].comment=-1;
   for(i=0;i<c->s.n-1;i++)
   {
     if (c->txt[i]=='\n')
     {
-      if (!flag) fputc('\n',fout); 
+      if (!flag) fputc('\n',fout);
       genindent(fout);
       while(c->txt[i+1]==' ' || c->txt[i+1]=='\t')
         i++;
@@ -56,19 +71,27 @@ void gencomment(FILE *fout, int n, int flag)
   }
   if (c->txt[i]=='\n')
   {
-    if (!flag) fputc('\n',fout); 
-    if (!(tree[n].m&(1<<Msemicomma)))
+    if (!flag) fputc('\n',fout);
+    if (!(tree[n].m&(1<<Msemicolon)))
       genindent(fout);
   }
   else
     fputc(c->txt[i],fout);
 }
-void gensemicomma(FILE *fout, int x)
+void gensemicolon(FILE *fout, int x)
 {
-  if (x!=-1 && !(tree[x].m&(1<<Msemicomma)))
+  if (x!=-1 && !(tree[x].m&(1<<Msemicolon)))
     fprintf(fout,";\n");
 }
 extern int optcleanvar;
+
+void genendbrace(FILE * fout, int x)
+{
+  genindentline(fout,"}");
+  if (!(tree[x].m&(1<<Muntil)))
+    fprintf(fout,"\n");
+}
+
 void genbrace(FILE *fout, int x)
 {
   if (x==-1) return;
@@ -76,192 +99,187 @@ void genbrace(FILE *fout, int x)
   {
   /*If it is an empty block, cross it*/
     if ((tree[x].m&(1<<Mbrace)) && !optcleanvar )
+    {
+      if ((tree[x].m)&(1<<Muntil))
+        tree[tree[x].y].m|=1<<Muntil;
       genbrace(fout,tree[x].y);
+    }
     else
-      gencodeg(fout,x);
+      gencode(fout,x);
   }
   else
   {
-    int brace=(tree[x].f==Fseq || (tree[x].m&(1<<Melse)));
+    int brace=(tree[x].f==Fseq || (tree[x].m&(1<<Melse)) || (tree[x].m&(1<<Muntil)));
     if (brace)
-    {
-      genindent(fout);
-      fprintf(fout,"{\n");
-    }
+      genindentline(fout,"{\n");
     indent++;
     genindentseq(fout,x);
-    gencodeg(fout,x);
-    gensemicomma(fout,x);
+    gencode(fout,x);
+    gensemicolon(fout,x);
     indent--;
     if (brace)
-    {
-      genindent(fout);
-      fprintf(fout,"}\n");
-    }
+      genendbrace(fout,x);
   }
 }
 
-void gensmallval(FILE *fout, int val, Gtype t)
+void genstring(FILE *fout, const char *s)
 {
-  switch(val)
+  fputc('"',fout);
+  while(*s)
   {
-    case 0:
-      fprintf(fout,(ctype[t]==Vsmall)?"0":"gzero");
-      break;
-    case 1:
-      fprintf(fout,(ctype[t]==Vsmall)?"1":"gun");
-      break;
-    case 2:
-      fprintf(fout,(ctype[t]==Vsmall)?"2":"gdeux");
-      break;
-    default:
-      die(-1,"Internal error:val %d unhandled in gensmallval",val);
+    if (*s=='\r')
+    {
+      s++;
+      continue;
+    }
+    if (*s=='\n')
+    {
+      fprintf(fout,"\\n\"\n");
+      genindent(fout);
+      fputc('"',fout);
+      s++;
+    }
+    else
+      fputc(*s++,fout);
   }
+  fputc('"',fout);
 }
-
 void genpercent(FILE *fout, int n)
 {
   const char *s=value[tree[n].x].val.str;
   while(*s)
   {
+    if (*s=='\r')
+    {
+      s++;
+      continue;
+    }
     if (*s=='%')
       fputc('%',fout);
-    fputc(*s++,fout);
-  }
-}
-
-void 
-genfacteurmat(FILE *fout, int n)
-{ 
-  /*
-     int stack[STACKSZ];
-     listtostackparent(n,Ffacteurmat,stack,0,STACKSZ,"indices",n);
-   */
-  int x=tree[n].x;
-  int y=tree[n].y;
-  int yx=tree[y].x;
-  int yy=tree[y].y;
-  int t=tree[x].t;
-  int f=tree[y].f;
-  enum {Tvec, Tvecsmall,Tlist} ty;
-  if (f!=Fmatrix && f!=FmatrixR)
-    die(y,"Internal error:genfacteurmat: unhandled func %s",funcname(f));
-  /* For now, we treat gen as vec, which will cause a bug if a gen 
-   * contain a vecsmall*/
-  if (t==Ggen || is_subtype(t,Gvec)) ty=Tvec;
-  else if (is_subtype(t,Gvecsmall)) ty=Tvecsmall;
-  else if(is_subtype(t,Glist)) ty=Tlist;
-  else
-    die(n,"Type %s has no components.",Gname[t]);
-  /*For vecsmall, we only allow [x], not [,x] nor [x,y]*/
-  if ((ty==Tvecsmall || ty==Tlist) && (f!=Fmatrix || yy!=-1))
-    die(n,"Type %s is not a matrix.",Gname[t]);
-  if (yy==-1)
-  {
-    genparens(fout,x);
-    fprintf(fout,"[");
-    gencast(fout,yx,(ty==Tlist)?Glg:Gsmall);
-    fprintf(fout,"]");
-  }
-  else
-  {
-    fprintf(fout,"coeff(");
-    gencast(fout,x,Ggen);
-    fprintf(fout,", ");
-    gencast(fout,yx,Gsmall);
-    fprintf(fout,", ");
-    gencast(fout,yy,Gsmall);
-    fprintf(fout,")");
+    if (*s=='\n')
+    {
+      fprintf(fout,"\\n\"\n");
+      genindent(fout);
+      fputc('"',fout);
+      s++;
+    }
+    else
+      fputc(*s++,fout);
   }
 }
 
 void genfacteuraff(FILE *fout, int x, int z)
 {
   int tx=tree[x].t;
-  int tz=tree[z].t;
-  gencodeg(fout,x);
+  if (x<0 || z<0) die(x,"Internal error: missing argument in =");
+  gencode(fout,x);
   fprintf(fout," = ");
-  if ((tree[x].m&(1<<Mlong)) && ctype[tx]==Vgen)
-  {
-    if (tree[z].f==Fentry && ctype[tz]==Vgen) 
-    {
-      fprintf(fout,"lcopy(");
-      gencast(fout,z,tx);
-      fprintf(fout,")");
-      return;
-    }
-    else
-      gencastl(fout,z,tx,0);
-  }
-  else
-    gencast(fout,z,tx);
+  gencast(fout,z,tx);
 }
 
-/*If there is no suitable GP prototype, just print nothing*/
 extern char *optsuffix;
-extern int optstrict;
-void geninstall(FILE *fout, int nf)
+void genaddhelp(FILE *fout, int nf)
+{
+  int stack[STACKSZ];
+  gpfunc *gp=lfunc+nf;
+  int h = gp->proto.help;
+  int nb, i;
+  if (h<0) return;
+  fprintf(fout,"GP;addhelp(%s%s, \"",
+      gp->gpname,(optsuffix?optsuffix:""));
+  nb=genlistcats(h,stack,STACKSZ);
+  for(i=0;i<nb;i++)
+  {
+    int d = stack[i];
+    const char *s;
+    if (tree[d].f!=Fentry && !is_const(d,CSTstr))
+       die(h,"Incorrect help text in addhelp");
+    s = entryname(d);
+    while(*s)
+    {
+      if (*s=='\n')
+        fputs("\\n",fout);
+      else if (*s!='\r')
+        fputc(*s,fout);
+      s++;
+    }
+  }
+  fputs("\");\n",fout);
+}
+
+int checkinstall(gpfunc *gp)
 {
   int i;
   int nargs=0;
+  context *fc=block+gp->user->bl;
+  if (!descfindrules1(gp->node, FC_proto_ret))
+    return 1;
+  for (i=0;i<fc->s.n;i++)
+  {
+    ctxvar *v=fc->c+i;
+    if ( v->flag&(1<<Carg) )
+    {
+      if (descfindrules1(v->node, FC_proto_code))
+        nargs++;
+      else
+        return 1;
+    }
+  }
+  if (max_args>0 && nargs>max_args)
+    return 1;
+  return 0;
+}
+
+/*If there is no suitable GP prototype, just print nothing*/
+void geninstall(FILE *fout, int nf)
+{
+  int i;
   gpfunc *gp=lfunc+nf;
-  int rt=ctype[gp->type];
-  if (!creturn[rt])
-    return;
   if(gp->spec==GPuser)
   {
     context *fc=block+gp->user->bl;
+    if (checkinstall(gp))
+      return;
+    fprintf(fout,"GP;install(\"%s\",\"",gp->proto.cname);
+    genfuncbydesc1(fout,gp->node,FC_proto_ret,-1);
     for (i=0;i<fc->s.n;i++)
     {
       ctxvar *v=fc->c+i;
       if ( v->flag&(1<<Carg) )
       {
-        int t=ctype[v->t];
-        if (cproto[t])
-          nargs++;
-        else
-          /*There are no suitable prototype*/
-	  return;
-      }
-    }
-    /*GP (<=2.2.3) cannot handle functions with more than 8 args*/
-    if (nargs>8 && optstrict)
-      return; 
-    fprintf(fout,"install(\"%s\",\"%s",gp->proto.cname,creturn[rt]);
-    for (i=0;i<fc->s.n;i++)
-    {
-      ctxvar *v=fc->c+i;
-      if ( v->flag&(1<<Carg) )
-      {
-        int t=ctype[v->t];
-        if (v->initval>=0 || !optstrict)
+        if (v->flag&(1<<Cvararg))
+          fprintf(fout,"s*");
+        else if (v->flag&(1<<Cdefmarker))
         {
-          fprintf(fout,"D");
-          if (t!=Vgen || !optstrict)
-          {
-            if (v->initval<0)
-              fprintf(fout,"%s",cdefault[t]);
-            else
-              printnode(fout,v->initval);
-            fprintf(fout,",%s,",cproto[t]);
-          }
-          else
-            fprintf(fout,"%s",cproto[t]);
+          fputc('D',fout);
+          genfuncbydesc1(fout,v->node,FC_proto_code,-1);
         }
+        else if (v->initval<0)
+          genfuncbydesc1(fout,v->node,FC_proto_code,-1);
         else
-          fprintf(fout,"%s",cproto[t]);
+        {
+          fputc('D',fout);
+          printnode(fout,v->initval);
+          fputc(',',fout);
+          genfuncbydesc1(fout,v->node,FC_proto_code,-1);
+          fputc(',',fout);
+        }
       }
     }
-    if (gp->mode&(1<<Mprec))
+    if (funcmode(*gp)&(1<<Mprec))
       fputc('p',fout);
   }
   else
-    fprintf(fout,"install(\"%s\",\"%s%s",gp->proto.cname,creturn[rt],gp->proto.code);
+  {
+    fprintf(fout,"GP;install(\"%s\",\"",gp->proto.cname);
+    genfuncbydesc1(fout,gp->node,FC_proto_ret,-1);
+    fprintf(fout,"%s",gp->proto.code);
+  }
   /* Member function has a gpname of "_.func" which is not a valid
    * GP func name.
    * we use "m_" instead.
    */
-  if( gp->gpname[0]=='_' && gp->gpname[1]=='.' ) 
+  if( gp->gpname[0]=='_' && gp->gpname[1]=='.' )
     fprintf(fout,"\",\"m_%s%s\"",gp->gpname+2,(optsuffix?optsuffix:""));
   else
     fprintf(fout,"\",\"%s%s\"",gp->gpname,(optsuffix?optsuffix:""));
@@ -269,24 +287,48 @@ void geninstall(FILE *fout, int nf)
     fprintf(fout,",\"%s\"",gp->proto.origin);
   fprintf(fout,");\n");
 }
+
+const char * genoutfile(void)
+{
+  if (outfile)
+  {
+    long n = strlen(outfile);
+    if (n > 2 && outfile[n-1]=='c' && outfile[n-2]=='.')
+    {
+      char *s = strdup(outfile);
+      s[n-2] = 0;
+      return s;
+    }
+  }
+  return strdup(nameparse);
+}
+
 void genheader(FILE *fout)
 {
   int i;
+  const char *file = genoutfile();
   fprintf(fout,"/*-*- compile-command: \"");
-  fprintf(fout,PARI_MODULE_BUILD,nameparse,nameparse,nameparse,nameparse);
+  fprintf(fout,PARI_MODULE_BUILD,file,file,file,file);
+  free((void*)file);
   fprintf(fout,"\"; -*-*/\n");
-  printf("#include <pari/pari.h>\n");
-  printf("/*\n");
+  fprintf(fout,"#include <pari/pari.h>\n");
+  fprintf(fout,"/*\n");
   for(i=0;i<s_func.n;i++)
     if (lfunc[i].spec==GPinstalled)
+    {
       geninstall(fout,i);
+      genaddhelp(fout,i);
+    }
   for(i=0;i<s_func.n;i++)
-    if (lfunc[i].spec==GPuser)
+    if (lfunc[i].spec==GPuser && !(lfunc[i].user->flag&(1<<UFstatic)))
+    {
       geninstall(fout,i);
-  printf("*/\n");
+      genaddhelp(fout,i);
+    }
+  fprintf(fout,"*/\n");
   for(i=0;i<s_func.n;i++)
   {
-    if (lfunc[i].spec==GPuser)
+    if (lfunc[i].spec==GPuser && !(lfunc[i].user->flag&(1<<UFstatic)))
       genprototype(fout,i,0);
     else if (lfunc[i].spec==GPinstalled)
       genprotocode(fout,i);
@@ -300,79 +342,74 @@ void genheader(FILE *fout)
     {
       ctxvar *v=ctxstack+i;
       fprintf(fout,"static ");
-      fprintf(fout,cname[ctype[v->t]],v->cvar);
+      genvarproto(fout,v->node,v->node);
       fprintf(fout,";\n");
     }
     fprintf(fout,"/*End of global vars*/\n\n");
   }
 }
 
-void gencodeg(FILE *fout, int n)
+void gencode(FILE *fout, int n)
 {
   int x=tree[n].x;
   int y=tree[n].y;
   if (n<0)
-    return; 
+    return;
   if (tree[n].comment>=0)
   {
-    if (tree[n].m&(1<<Msemicomma))
+    if (tree[n].m&(1<<Msemicolon))
       genindent(fout);
     gencomment(fout,n,0);
   }
-  if (debug>=3) fprintf(fout,"/*%s:%d*/",Gname[tree[n].t],tree[n].m);
+  if (debug>=3) fprintf(fout,"/*%s:%d*/",GPname(tree[n].t),tree[n].m);
   switch(tree[n].f)
   {
   case Fseq:
     genindentseq(fout,x);
-    gencodeg(fout,x);
-    gensemicomma(fout,x);
+    gencode(fout,x);
+    gensemicolon(fout,x);
     genindentseq(fout,y);
-    gencodeg(fout,y);
-    gensemicomma(fout,y);
+    gencode(fout,y);
+    gensemicolon(fout,y);
     break;
-  case Fmatrix:
-  case FmatrixL:
-  case FmatrixR:
-    die(-1,"Internal error: Fmatrix* in gencode");
-    /*all this stuff is handled by Ffacteurmat*/
-    break;
-  case Ffacteurmat:
-    genfacteurmat(fout,n);
-    break;
-  case Faffect:
+  case Fassign:
     genfacteuraff(fout,x,y);
     break;
   case Fconst:
     {
       long val=value[x].val.small;
       const char *str=value[x].val.str;
-      switch (value[x].type)
+      long typ=value[x].type;
+      switch (typ)
       {
       case CSTsmall:
         fprintf(fout,"%ld",val);
         break;
       case CSTsmallreal:
-        switch(val)
         {
-        case 0:
-          fprintf(fout,"realzero(prec)");
-          break;
-        case 1:
-          fprintf(fout,"realun(prec)");
-          break;
-        default:
-          fprintf(fout,"mulsr(%ld, realun(prec))",val);
-          break;
+          int arg=newsmall(val);
+          tree[arg].t=Gsmall;
+          if (genfuncbydesc1(fout,arg,FC_const_smallreal,n))
+            die(n,"Bad description for _const_smallreal");
+          stack_pop_safe(&s_node,arg);
         }
         break;
-      case CSTexpr:
-        fprintf(fout,"flisexpr(\"%s\")",str);
+      case CSTint:
+      case CSTreal:
+        {
+          int arg=newstringnode(str,-1);
+          int fc=FC_const_expr;
+          tree[arg].t=Gstr;
+          if (FC_const_real>=0 && typ==CSTreal)
+            fc=FC_const_real;
+          if (genfuncbydesc1(fout,arg,fc,n))
+            die(n,"Bad description for _const_expr");
+          stack_pop_safe(&s_value,tree[arg].x);
+          stack_pop_safe(&s_node,arg);
+        }
         break;
       case CSTstr:
-        fprintf(fout,"\"%s\"",str);
-        break;
-      case CSTquote:
-        fprintf(fout,"fetch_user_var(\"%s\")",str);
+        genstring(fout,str);
         break;
       }
       break;
@@ -380,25 +417,19 @@ void gencodeg(FILE *fout, int n)
   case Fsmall:
     fprintf(fout,"%d",tree[n].x);
     break;
-  case Femptyvec:
-    fprintf(fout,"cgetg(1,t_VEC)");
-    break;
-  case Femptymat:
-    fprintf(fout,"cgetg(1,t_MAT)");
-    break;
-  case Fmatrixelts:
-  case Fmatrixlines:
-  case Fmat:
-    /*handled by genblockmatrix*/
-  case Flistarg:
-    /*handle by Fentry*/
-    die(-1,"Internal error: %s in gencode",funcname(tree[n].f));
-    break;
   case Frefarg:
+    gencode(fout,x);
+    break;
   case Fentry:
     genentry(fout,n);
     break;
-  case Fentryfunc:
+  case Ffun:
+    {
+      gpfunc *gp=lfunc+tree[n].x;
+      fprintf(fout,"%s",gp->proto.cname);
+    }
+    break;
+  case Ffunction:
     genentryfunc(fout,n);
     break;
   case Fdeffunc:
@@ -407,14 +438,14 @@ void gencodeg(FILE *fout, int n)
   case Fblock:
     gendefblock(fout,n);
     break;
-  case Fgnil:
-    fprintf(fout,"0");
+  case Fnoarg:
+    fputc('0',fout);
     break;
   case Ftag:
     gencast(fout,x,y);
     break;
   default:
-    die(n,"Internal error : unknow func %s in gencode",funcname(tree[n].f));
+    die(n,"Internal error : unknown func %s in gencode",funcname(tree[n].f));
     break;
   }
 }
@@ -422,26 +453,16 @@ void genparensg(FILE *fout, int n)
 {
   if (n!=-1 && (tree[n].m&(1<<Mparens)))
     fprintf(fout,"(");
-  gencodeg(fout, n);
+  gencode(fout, n);
   if (n!=-1 && (tree[n].m&(1<<Mparens)))
     fprintf(fout,")");
 }
-void gencode(FILE *fout, int n)
-{
-  if ((tree[n].m&(1<<Mlong)) && ctype[tree[n].t]==Vgen)
-  {
-    fprintf(fout,"(GEN) ");
-    genparensg(fout, n);
-  }
-  else
-    gencodeg(fout, n);
-}
 void genparens(FILE *fout, int n)
 {
   long x=n;
   if(n<0) return;
   x=detag(x);
-  if (tree[x].m&(1<<Mlong|1<<Mparens))
+  if (tree[x].m&(1<<Mparens))
   {
     fprintf(fout,"(");
     gencode(fout, n);
